
#ifndef PLS_SCHEDULER_H
#define PLS_SCHEDULER_H

#include <array>
#include <iostream>

#include "pls/internal/base/aligned_stack.h"
#include "pls/internal/base/thread.h"
#include "pls/internal/base/barrier.h"

#include "thread_state.h"
#include "root_task.h"

namespace pls {
    namespace internal {
        namespace scheduling {
            // Upper thread limit for static memory allocation.
            // Could be moved to templating if needed.
            static constexpr int MAX_THREADS = 32;

            void worker_routine();
            using scheduler_thread = base::thread<decltype(&worker_routine), thread_state>;

            class scheduler_memory {
            public:
                virtual size_t max_threads() = 0;
                virtual thread_state* thread_state_for(size_t id) = 0;
                virtual scheduler_thread* thread_for(size_t id) = 0;
                virtual base::aligned_stack* task_stack_for(size_t id) = 0;
            };

            template<size_t MAX_THREADS, size_t TASK_STACK_SIZE>
            class static_scheduler_memory: public scheduler_memory {
                std::array<scheduler_thread, MAX_THREADS> threads_;
                std::array<thread_state, MAX_THREADS> thread_states_;
                std::array<std::array<char, TASK_STACK_SIZE>, MAX_THREADS> task_stacks_memory_;
                std::array<base::aligned_stack, MAX_THREADS> task_stacks_;

            public:
                static_scheduler_memory() {
                    for (size_t i = 0; i < MAX_THREADS; i++) {
                        task_stacks_[i] = base::aligned_stack(reinterpret_cast<char*>(&task_stacks_memory_[i]), TASK_STACK_SIZE);
                    }
                }

                size_t max_threads() override { return MAX_THREADS; }
                thread_state* thread_state_for(size_t id) override { return &thread_states_[id]; }
                scheduler_thread* thread_for(size_t id) override { return &threads_[id]; }
                base::aligned_stack* task_stack_for(size_t id) override { return &task_stacks_[id]; }
            };

            class scheduler {
                friend void worker_routine();

                const unsigned int num_threads_;
                scheduler_memory* memory_;

                base::barrier sync_barrier_;
                bool terminated_;
            public:
                explicit scheduler(scheduler_memory* memory, unsigned int num_threads);
                ~scheduler();

                template<typename Function>
                void perform_work(Function work_section) {
                    root_task<Function> master{work_section};

                    // Push root task on stacks
                    auto new_master = memory_->task_stack_for(0)->push(master);
                    memory_->thread_state_for(0)->root_task_ = new_master;
                    memory_->thread_state_for(0)->current_task_ = new_master;
                    for (unsigned int i = 1; i < num_threads_; i++) {
                        root_worker_task<Function> worker{new_master};
                        auto new_worker = memory_->task_stack_for(0)->push(worker);
                        memory_->thread_state_for(i)->root_task_ = new_worker;
                        memory_->thread_state_for(i)->current_task_ = new_worker;
                    }

                    // Perform and wait for work
                    sync_barrier_.wait(); // Trigger threads to wake up
                    sync_barrier_.wait(); // Wait for threads to finish

                    // Clean up stack
                    memory_->task_stack_for(0)->pop<typeof(master)>();
                    for (unsigned int i = 1; i < num_threads_; i++) {
                        root_worker_task<Function> worker{new_master};
                        memory_->task_stack_for(0)->pop<typeof(worker)>();
                    }
                }

                // TODO: See if we should place this differently (only for performance reasons)
                template<typename Task>
                static void execute_task(Task& task, int depth=-1) {
                    static_assert(std::is_base_of<abstract_task, Task>::value, "Only pass abstract_task subclasses!");

                    auto my_state = base::this_thread::state<thread_state>();
                    abstract_task* old_task;
                    abstract_task* new_task;

                    // Init Task
                    {
                        std::lock_guard<base::spin_lock> lock{my_state->lock_};
                        old_task = my_state->current_task_;
                        new_task = my_state->task_stack_->push(task);

                        new_task->set_depth(depth >= 0 ? depth : old_task->depth() + 1);
                        my_state->current_task_ = new_task;
                        old_task->set_child(new_task);
                    }

                    // Run Task
                    new_task->execute();

                    // Teardown state back to before the task was executed
                    {
                        std::lock_guard<base::spin_lock> lock{my_state->lock_};

                        old_task->set_child(nullptr);
                        my_state->current_task_ = old_task;

                        my_state->task_stack_->pop<Task>();
                    }
                }

                void terminate(bool wait_for_workers=true);
                unsigned int num_threads() const { return num_threads_; }
                thread_state* thread_state_for(size_t id) { return memory_->thread_state_for(id); }
            };
        }
    }
}

#endif //PLS_SCHEDULER_H
